#
# (c) 2022, Giorgio Gonnella, University of Goettingen, Germany
#

"""
Helper methods for working with the database
"""

import os
from sqlalchemy.engine.url import URL
from schema import And, Or

ARGS_DOC = """\
  dbuser:       database user to use
  dbpass:       password of the database user
  dbname:       database name
  dbsocket:     connection socket file\
"""

OPTS_DOC = """\
  --dbpfx PFX      database tablenames prefix to use (default: prenacs_)\
"""

ARGS_USAGE="<dbuser> <dbpass> <dbname> <dbsocket>"

ARGS_SCHEMA = {"<dbuser>": And(str, len),
               "<dbpass>": And(str, len),
               "<dbname>": And(str, len),
               "<dbsocket>": And(str, len, os.path.exists),
               "--dbpfx": Or(None, And(str, len))}

SNAKE_ARGS = {"config": ["<dbuser>", "<dbpass>", "<dbname>"],
              "params": ["--dbpfx"],
              "input": ["<dbsocket>"]}

DB_DRIVER="mysql+mysqldb"
DB_HOST="localhost"

def connection_string_from(args) -> str:
  """
  MySQL/MariaDB connection string based on the values of the args
  '<dbuser>', '<dbpass>', '<dbname>' and '<dbsocket>'
  """
  if not(args["<dbsocket>"]):
    raise RuntimeError("DB unix socket not provided")
  elif not os.path.exists(args["<dbsocket>"]):
    raise RuntimeError(f"DB unix socket does not exist: {args['<dbsocket>']}")
  elif not(args["<dbuser>"]):
    raise RuntimeError("Database user name not provided")
  elif not(args["<dbpass>"]):
    raise RuntimeError("Database password for user '{}' not provided".\
                      format(args["<dbuser>"]))
  elif not(args["<dbname>"]):
    raise RuntimeError("Database name not provided")
  return URL.create(drivername=DB_DRIVER,
                    username=args["<dbuser>"],
                    password=args["<dbpass>"],
                    database=args["<dbname>"],
                    host=DB_HOST,
                    query={"unix_socket":args["<dbsocket>"]})



#
# (c) 2021-2022 Giorgio Gonnella, University of Goettingen, Germany
#
from sqlalchemy import select
from sqlalchemy.orm import Session
from prenacs.dbschema.attribute_definition import AttributeDefinition

class AttributeDefinitionsManager():

  def __init__(self, attribute_value_tables):
    self.avt = attribute_value_tables
    self.connection = attribute_value_tables.connectable

  def _check_invariant_column(self, adef, fname, fvalue):
    prev_value = getattr(adef, fname)
    if prev_value != fvalue:
      raise RuntimeError("Cannot update the definition of "+\
          f"attribute '{adef.name}' because changes in the column "+\
          f"{fname} are not allowed (previous value: "+\
          f"{prev_value}, current value: {fvalue})")

  def update(self, name, definition):
    """
    Update the definition of an attribute.

    The new definition must contain the same value as the previous definition
    for the keys "datatype" and "computation_group".

    All fields except "name", "datatype" and "computation_group" are updated.

    :param name: the name of the attribute
    :param definition: the new definition of the attribute, as a dictionary
    """
    session = Session(self.connection)
    select_adefs = select(AttributeDefinition).where(\
        AttributeDefinition.name == name)
    adef = session.execute(select_adefs).scalars().first()
    if adef is None:
      raise RuntimeError("Cannot update the definition of attribute "+\
          f"'{name}' because it does not exist")
    for fname, fvalue in definition.items():
      if fname in ["datatype", "computation_group"]:
        self._check_invariant_column(adef, fname, fvalue)
      else:
        setattr(adef, fname, fvalue)
    session.add(adef)
    session.commit()

  def update_changed(self, definitions):
    """
    Update the definitions of attributes given in a dictionary
    that are different from the values in the database.

    :param definitions: dictionary containing the attribute name as keys and the
                        other definition columns as values

    If an attribute given in the dictionary does not exist in the
    database, or if exists, but it is unchanged, it is ignored.
    """
    if definitions:
      session = Session(self.connection)
      select_adefs = select(AttributeDefinition)
      for adef in session.execute(select_adefs).scalars().all():
        if adef.name in definitions:
          for fname, fvalue in definitions[adef.name].items():
            if fname in ["datatype", "computation_group"]:
              self._check_invariant_column(adef, fname, fvalue)
            else:
              setattr(adef, fname, fvalue)
          session.add(adef)
      session.commit()

  def drop(self, name):
    """
    Drop an attribute from the database.

    :param name: the name of the attribute to drop
    """
    self.avt.destroy_attribute(name)

  def drop_missing(self, definitions):
    """
    Drop attributes that are not in the definitions.

    :param definitions: dictionary containing the attribute name as keys and the
                        other definition columns as values

    If an attribute present in the database does not exist in the
    definitions dictionary, it is removed from the database.

    Note that all values for the attribute are lost!
    """
    for aname in self.avt.attribute_names:
      if not definitions or aname not in definitions:
        self.avt.destroy_attribute(aname)

  def insert(self, name, definition):
    """
    Insert a new attribute into the database.

    :param name: the name of the attribute to insert
    :param definition: the definition of the attribute, as a dictionary
    """
    self.avt.create_attribute(name, **definition)

  def insert_new(self, definitions):
    """
    Insert in the database attributes from the definitions
    that are not yet in the database.

    :param definitions: dictionary containing the attribute name as keys and the
                        other definition columns as values

    If an attribute already exists in the database, it is ignored.
    """
    if definitions:
      for aname, adef in definitions.items():
        if aname not in self.avt.attribute_names:
          self.avt.create_attribute(aname, **adef)

  def apply_definition(self, aname, definition):
    if aname not in self.avt.attribute_names:
      self.avt.create_attribute(aname, **definition)
    else:
      self.update(aname, definition)

  def apply_definitions(self, definitions, drop_missing=True,
                        insert_new=True, update_changed=True):
    if drop_missing:   self.drop_missing(definitions)
    if insert_new:     self.insert_new(definitions)
    if update_changed: self.update_changed(definitions)



#
# (c) 2022, Giorgio Gonnella, University of Goettingen, Germany
#

from schema import Or, And, Use, Optional, Schema

COLNUM_VALIDATOR = And(Use(int), lambda n: n>0)
OPTCOLNUM_VALIDATOR = Or(And(None, Use(lambda n: 1)), COLNUM_VALIDATOR)

ARGS_DOC = """\
  --verbose, -v    be verbose
  --version, -V    show script version
  --help, -h       show this help message"""

ARGS_SCHEMA = {Optional("--verbose", default=None): Or(None, True, False),
                                                    Optional(str): object}

def validate(args, *dicts):
  s = ARGS_SCHEMA.copy()
  for d in dicts:
    s.update(d)
  return Schema(s).validate(args)



#
# (c) 2021-2022 Giorgio Gonnella, University of Goettingen, Germany
#

import yaml
import uuid
from datetime import datetime
import sys
import socket
import getpass

class Report():

  REASONS = ["new_entities", "new_attributes", "recompute"]

  def _init_plugin(self, plugin):
    if plugin is None:
      raise ValueError("A plugin module must be provided for creating a report")
    if not hasattr(plugin, "ID"):
      raise ValueError("The plugin module must have an ID attribute")
    if not hasattr(plugin, "VERSION"):
      raise ValueError("The plugin module must have a VERSION attribute")
    self.data["plugin_id"] = plugin.ID
    self.data["plugin_version"] = plugin.VERSION

  def _init_system(self, system):
    if system is None:
      system = socket.gethostname()
    self.data["system_id"] = system

  def _init_user(self, user):
    if user is None:
      user = getpass.getuser()
    self.data["user_id"] = user

  def _init_reason(self, reason):
    if reason is not None:
      if reason not in self.REASONS:
        raise ValueError("Invalid value for the computation reason; "+\
            "it must be one of: " + str(self.REASONS))
    self.data["reason"] = reason

  def _init_params(self, params):
    if params is not None:
      self.data["parameters"] = yaml.dump(params)

  def __init__(self, reportfile, plugin, user = None,
               system = None, reason = None, params = {}):
    self.data = {}
    self.rfile = reportfile
    self._init_plugin(plugin)
    self._init_system(system)
    self._init_user(user)
    self._init_reason(reason)
    self._init_params(params)
    self.data["uuid"] = uuid.uuid4().bytes
    self.data["time_start"] = str(datetime.now())
    self.n_steps = 0

  def step(self):
    self.n_steps += 1

  def finalize(self):
    self.data["time_end"] = str(datetime.now())
    self.data["n_units"] = self.n_steps
    self.data["comp_status"] = "completed"
    yaml.dump(self.data, self.rfile)
    self.rfile.flush()
    if self.rfile != sys.stderr:
      self.rfile.close()

  def error(self, err, unitname):
    self.data["time_end"] = str(datetime.now())
    self.data["n_units"] = self.n_steps
    self.data["comp_status"] = "aborted" if self.n_steps == 0 else "partial"
    remark = {}
    remark["error_input_unit"] = unitname
    remark["error_class"] = err.__class__.__name__
    remark["error_message"] = str(err)
    self.data["remarks"] = yaml.dump(remark)
    yaml.dump(self.data, self.rfile)
    self.rfile.flush()
    if self.rfile != sys.stderr:
      self.rfile.close()




#
# (c) 2021-2022 Giorgio Gonnella, University of Goettingen, Germany
#
def shorten(s: str, limit: int, placeholder: str = "...") -> str:
  """
  Shorten a string to a string with maximum length limit.

  If the string is longer than limit, the string is shortened
  to limit - len(placeholder) character and the placeholder is added.
  """
  assert(limit >= len(placeholder))
  return s if len(s) <= limit else s[:limit-len(placeholder)] + placeholder



#
# (c) 2021-2022 Giorgio Gonnella, University of Goettingen, Germany
#
import sys
from pathlib import Path
from glob import glob
from prenacs import plugins_helper, formatting_helper
from prenacs.report import Report
import tqdm
from concurrent.futures import as_completed, ProcessPoolExecutor
import multiplug

plugin=None

class EntityProcessor():
  @staticmethod
  def run(input_id, params):
    global plugin
    return plugin.compute(input_id, **params)

class BatchComputation():
  def __init__(self, plugin, verbose=False):
    """
    Initialize a batch computation instance based on a plugin.
    """
    self.plugin = multiplug.importer(plugin, verbose=verbose,
                                     **plugins_helper.COMPUTE_PLUGIN_INTERFACE)
    self.desc = formatting_helper.shorten(Path(plugin).stem, 15)
    self.state = None
    self.outfile = sys.stdout
    self.logfile = sys.stderr
    self.all_ids = None
    self.report = None
    self.params = {}
    self.computed = False

  def _compute_skip_set(self, skip_arg, verbose):
    skip = set()
    if skip_arg:
      if verbose:
        sys.stderr.write(f"# processing skip list... ({skip_arg})\n")
      with open(skip_arg) as f:
        for line in f: skip.add(line.rstrip().split("\t")[0])
      if verbose:
        sys.stderr.write("# done: skipping computation for "+\
                         f"up to {len(skip)} units\n")
    elif verbose:
      sys.stderr.write("# no skip list, all input units will be processed\n")
    return skip

  def _get_mod_function(self, filename, fun, verbose):
    if filename:
      pmod = multiplug.importer(filename, verbose=verbose,
                                **plugins_helper.IDPROC_PLUGIN_INTERFACE)
      return getattr(pmod, fun)
    else:
      return None

  def _compute_ids(self, unit_name, is_filename, idsproc):
    identifier = idsproc(unit_name) if idsproc else unit_name
    return (unit_name if is_filename else identifier), identifier

  def _input_units(self, globpattern, idsfile, idscol):
    if globpattern:
      return glob(globpattern)
    else:
      with open(idsfile) as f:
        return [line.rstrip().split("\t")[idscol-1] for line in f]

  def _compute_all_ids(self, globpattern, idsfile, idscol,
                       idsproc, skip, verbose):
    if verbose:
      sys.stderr.write("# compute input and output IDs...")
    result = []
    for unit_name in self._input_units(globpattern, idsfile, idscol):
      input_id, output_id = self._compute_ids(unit_name, globpattern, idsproc)
      if output_id in skip:
        skip.remove(output_id)
      else:
        result.append((input_id, output_id))
    if verbose:
      sys.stderr.write("done\n")
    return result

  def _select_input(self, globpattern=None, idsfile=None, idscol=None,
                  idsproc_module=None, skip=None, verbose=False):
    idsproc = self._get_mod_function(idsproc_module, "compute_id", verbose)
    skip = self._compute_skip_set(skip, verbose)
    self.all_ids = self._compute_all_ids(globpattern, idsfile, idscol,
                                         idsproc, skip, verbose)

  def input_from_globpattern(self, globpattern, idsproc_module=None,
                             skip=None, verbose=False):
    """
    Select the computation input using a glob pattern.

    # Input IDs

    The filename of each file in the glob pattern is used as the input
    ID passed to the plugin compute function.

    # Output IDs

    If a idsproc_module is specified, the ``compute_id`` function of
    the module is used to compute the output ID.
    Otherwise the filename is used also as the output ID.

    # Skip computation for some units

    A skip list file can be be defined using the parameter ``skip``.
    If skip is defined, then the computation is skipped for any
    output ID that is in the skip list.

    The skip list file can:
    - contain one output ID per line, or
    - be a tab-separated file with the output ID in the first column.
    """
    self._select_input(globpattern, None, None, idsproc_module, skip, verbose)

  def input_from_idsfile(self, idsfilename, idscol=1, idsproc_module=None,
                         skip=None, verbose=False):
    """
    Select the computation input using a tab-separated file.

    # Input IDs

    The entity IDs are passed as input to the plugin compute function.
    The IDs are obtained from the column specified by the parameter
    ``idscol`` (1-based column number) of the tab-separated file
    ``idsfilename``.

    If a idsproc_module is specified, the ``compute_id`` function of
    the module is used to compute the IDs from the column values.
    Otherwise the IDs are the column values themselves.

    # Output IDs

    The same IDs are used as input and output IDs.

    # Skip computation for some units

    A skip list file can be be defined using the parameter ``skip``.
    If skip is defined, then the computation is skipped for any
    ID that is in the skip list.

    The skip list file can:
    - contain one ID per line, or
    - be a tab-separated file with the ID in the first column.
    """
    if idscol < 1:
      raise ValueError("idscol must be a positive integer")
    self._select_input(None, idsfilename, idscol, idsproc_module, skip, verbose)

  def set_output(self, outfilename = None, logfilename = None):
    self.outfile = open(outfilename, "a") if outfilename else sys.stdout
    self.logfile = open(logfilename, "a") if logfilename else sys.stderr

  def setup_computation(self, params = {}, reportfile = sys.stderr,
                        user = None, system = None, reason = None,
                        verbose = False):
    """
    Setup the computation.

    It allows to set computation and report parameters and run
    the plugin initialization code.

    # Computation parameters

    Computation parameters are passed as a dictionary (``params``):
    - ``state``: a state object, which can change during the computation;
      it is passed to and/or setup by the initialization function of the plugin
      (see below)
    - any other key: any other parameter that the plugin may need
      and remains constant during the computation

    If no parameters are needed, then params can be omitted and defaults
    to an empty dictionary.

    ## Plugin initialization

    If the plugin has a ``initialize`` method, it is called.
    The return value of the ``initialize`` method is stored in the
    ``state`` parameter.

    If the parameter ``state`` was already defined, it is passed to the
    ``initialize`` method, and then overwritten with the return value of the
    ``initialize`` method.

    If a plugin initialization is not needed, then the plugin shall
    not define an ``initialize`` method.

    ## Report

    The following computation report parameters can be set:
    - ``reportfile`` sets the output file for the report (default: stderr)
    - ``user`` sets the user name for the report (default: system user
       as determined using ``getpass.getuser()``)
    - ``system`` sets the system name for the report (default: system
       name as determined using ``socket.gethostname()``)
    - ``reason`` sets the reason for the computation (default: undefined);
       it must be None or one of the reasons given in Report.REASONS

    If no report parameters are set, then the report is generated using
    the default values mentioned above.
    """
    if self.report:
      raise ValueError("Computation already set up")
    self.report = Report(reportfile, self.plugin,
                         user, system, reason, params)
    self.params = params
    if self.plugin.initialize is not None:
      self.params["state"] = \
          self.plugin.initialize(**self.params.get("state", {}))

  def _default_computation_setup(self):
    self.report = Report(sys.stderr, self.plugin)
    if self.plugin.initialize is not None:
      self.params["state"] = self.plugin.initialize()

  def _on_failure(self, output_id, exc):
    self.outfile.flush()
    self.logfile.flush()
    self.report.error(exc, output_id)

  def _on_success(self, output_id, results, logs):
    results = "\t".join([str(r) for r in results])
    if results:
      self.outfile.write(f"{output_id}\t{results}\n")
    for element in logs:
      if isinstance(element, list):
        for subelement in element:
          if subelement:
            self.logfile.write(f"{output_id}\t{subelement}\n")
      elif element:
        self.logfile.write(f"{output_id}\t{element}\n")
    self.report.step()

  def run(self, parallel=True, verbose=False):
    """
    Run the computation.

    By default, the computation is run in parallel
    (using a multiprocessing.Pool).
    Set parallel to False to run the computation serially.
    """
    if not self.all_ids:
      raise ValueError("Input was not selected")
    if not self.report:
      self._default_computation_setup()
    if parallel:
      self._run_in_parallel(verbose)
    else:
      self._run_serially(verbose)
    self.computed = True

  def _run_in_parallel(self, verbose):
    global plugin
    plugin = self.plugin
    entity_processor = EntityProcessor()
    if verbose:
      sys.stderr.write("# Computation will be in parallel (multiprocess)\n")
    with ProcessPoolExecutor() as executor:
      futures_map = {executor.submit(entity_processor.run,
                                     unit_ids[0], self.params):
                     unit_ids[1] for unit_ids in self.all_ids}
      for future in tqdm.tqdm(as_completed(futures_map), total=len(futures_map),
                              desc=self.desc):
        output_id = futures_map[future]
        try:
          results, *logs = future.result()
        except Exception as exc:
          self._on_failure(output_id, exc)
          raise(exc)
        else:
          self._on_success(output_id, results, logs)
    plugin = None

  def _run_serially(self, verbose):
    if verbose:
      sys.stderr.write("# Computation will be serial\n")
    for unit_ids in tqdm.tqdm(self.all_ids, desc=self.desc):
      output_id = unit_ids[1]
      try:
        results, *logs = self.plugin.compute(unit_ids[0], **self.params)
      except Exception as exc:
        self._on_failure(output_id, exc)
        raise(exc)
      else:
        self._on_success(output_id, results, logs)

  def finalize(self):
    """
    This method is called after the computation is finished.

    It finalizes the report, runs the plugin finalization code
    (if any) and closes the output files.
    """
    if not self.computed:
      raise ValueError("Computation not run")
    self.report.finalize()
    if self.plugin.finalize is not None:
      self.plugin.finalize(self.params.get("state", None))
    if self.outfile != sys.stdout: self.outfile.close()
    if self.logfile != sys.stderr: self.logfile.close()




#
# (c) 2021-2022 Giorgio Gonnella, University of Goettingen, Germany
#
"""
Analyses the interface of a computation plugin for compliance
with the interface specification.
"""

import sys
import inspect
import multiplug
from loguru import logger

class PluginInterfaceAnalyser():

  def __init__(self, plugin, verbose=True, outfile=sys.stdout, colorize=True):
    self.plugin = multiplug.importer(plugin, verbose=verbose)
    self.had_errors = False
    log_level = "INFO" if verbose else "WARNING"
    logger.add(outfile, colorize=colorize, level=log_level)

  def _error(self, msg):
    logger.error(msg)
    self.had_errors = True

  def _info(self, msg):
    logger.info(msg)

  def _success(self, msg):
    logger.success(msg)

  def _check_compute_function(self):
    if not hasattr(self.plugin, "compute"):
      self._error("plugin does not provide a compute function")
    else:
      self._success("plugin provides a compute function")
    if self.plugin.__lang__ == "python":
      compute_spec = inspect.getfullargspec(self.plugin.compute)
      if compute_spec.varargs is not None:
        self._error(\
            "plugin.compute() accepts variable positional arguments")
      if not hasattr(self.plugin, "initialize"):
        if len(compute_spec.args) != 1:
          self._error(\
              "plugin.compute() does not accept a single positional argument")
      else:
        if (len(compute_spec.args) != 2) or (compute_spec.args[1] != "state"):
          self._error(\
              "plugin.compute() in plugin with initialize function "+\
              "does not accept a state keyword argument")
      if compute_spec.varkw is None:
        self._error(\
            "plugin.compute() does not accept variable keywords arguments")

  def _check_initialize_function(self):
    if hasattr(self.plugin, "initialize"):
      self._success("plugin provides an initialize function")
    else:
      self._info("plugin does not provide an initialize function")
    init_spec = inspect.getfullargspec(self.plugin.initialize)
    if (init_spec.varargs is not None) or (len(init_spec.args) > 0):
      self._error(\
          "plugin.initialize() accepts positional arguments")
    if init_spec.varkw is None:
      self._error(\
          "plugin.initialize() does not accept variable keywords arguments")

  def _check_finalize_function(self):
    if hasattr(self.plugin, "finalize"):
      self._success("plugin provides a finalize function")
      if not hasattr(self.plugin, "initialize"):
        self._error(\
              "plugin.finalize() defined in plugin without plugin.initialize()")
    else:
      self._info("plugin does not provide a finalize function")
    f_spec = inspect.getfullargspec(self.plugin.finalize)
    if f_spec.varargs is not None:
      self._error(\
          "plugin.finalize() accepts variable positional arguments")
    if len(f_spec.args) != 1:
      self._error(\
          "plugin.finalize() does not accept a single positional argument")
    if f_spec.varkw is not None:
      self._error(\
          "plugin.initialize() accepts variable keywords arguments")

  def _check_mandatory_constants(self):
    for const in ["ID", "VERSION", "INPUT", "OUTPUT"]:
      if not hasattr(self.plugin, const):
        self._error(\
            f"plugin does not define a constant named '{const}'")
      else:
        self._success(f"plugin defines a constant named '{const}'")

  def _check_string_constants(self):
    for const, maxlen in {"ID": 256, "VERSION": 64, "INPUT": 512,
                          "METHOD": 4096, "IMPLEMENTATION": 4096,
                          "REQ_SOFTWARE": 4096, "REQ_HARDWARE": 4096,
                          "ADVICE": 4096}.items():
      if hasattr(self.plugin, const):
        v = getattr(self.plugin, const)
        if v is not None:
          if not isinstance(v, str):
            self._error(f"plugin.{const} must be a string")
          elif len(v) > maxlen:
            self._error(f"plugin.{const} length must <= {maxlen}")
          else:
            self._success(f"plugin.{const} type and format is valid")

  def _check_output_constant(self, definitions):
    if hasattr(self.plugin, "OUTPUT"):
      if not isinstance(self.plugin.OUTPUT, list):
        self._error("plugin.OUTPUT is not a list")
      notstr = [key for key in self.plugin.OUTPUT if not isinstance(key, str)]
      if notstr:
        self._error(\
            f"plugin.OUTPUT non-string elements found: {notstr}")
      else:
        self._success("all plugin.OUTPUT elements are strings")
      if definitions:
        notin = [key for key in self.plugin.OUTPUT \
                  if key not in notstr and key not in definitions]
        if notin:
          self._error("plugin.OUTPUT elements "+\
              f"not found in the attributes definitions file: {notin}")
        else:
          self._success("all plugin.OUTPUT elements are "+\
                "present in the attributes definitions file")
      else:
        self._info("no attributes definitions file provided")

  def _check_parameters_constant(self):
    if hasattr(self.plugin, "PARAMETERS") and \
        self.plugin.PARAMETERS is not None:
      if not isinstance(self.plugin.PARAMETERS, list):
        self._error("plugin.PARAMETERS is not a list")
      for element in self.plugin.PARAMETERS:
        if not isinstance(element, tuple) and not isinstance(element, list):
          self._error(\
              f"plugin.PARAMETER element not tuple: {element}")
        elif len(element) != 4:
          self._error(\
              f"plugin.PARAMETER element tuple length is not 4: {element}")
        else:
          for s in element:
            if not isinstance(s, str):
              self._error(\
                "plugin.PARAMETER tuple element is not a "+\
                f"string: {s} of {element}")

  def run(self, definitions=None):
    if self.plugin.__lang__ != "python":
      self._info("signature of plugin functions not analyzed, "+\
          " since it is a {} plugin", self.plugin.__lang__)
    self._check_compute_function()
    self._check_initialize_function()
    self._check_finalize_function()
    self._check_mandatory_constants()
    self._check_string_constants()
    self._check_output_constant(definitions)
    self._check_parameters_constant()
    return 1 if self.had_errors else 0




#
# (c) 2021-2022 Giorgio Gonnella, University of Goettingen, Germany
#
COMPUTE_PLUGIN_INTERFACE = {}
COMPUTE_PLUGIN_INTERFACE["req_func"] = ["compute"]
COMPUTE_PLUGIN_INTERFACE["opt_func"] = ["initialize", "finalize"]
COMPUTE_PLUGIN_INTERFACE["req_const"] = ["ID", "VERSION", "INPUT", "OUTPUT"]
COMPUTE_PLUGIN_INTERFACE["opt_const"] = ["PARAMETERS", "METHOD",
                                         "IMPLEMENTATION", "ADVICE",
                                         "REQ_SOFTWARE", "REQ_HARDWARE"]

IDPROC_PLUGIN_INTERFACE = {}
IDPROC_PLUGIN_INTERFACE["req_func"] = ["compute_id"]

def plugin_metadata_str(plugin):
  metadata_keys = COMPUTE_PLUGIN_INTERFACE["req_const"] + \
                  COMPUTE_PLUGIN_INTERFACE["opt_const"]
  result = {k.lower(): getattr(plugin, k) \
              for k in metadata_keys if hasattr(plugin, k)}
  if "output" in result and result["output"] is not None:
    result["output"] = ",".join(result["output"])
  if "parameters" in result and result["parameters"] is not None:
    result["parameters"] = ";".join([",".join(e) for e in result["parameters"]])
  return result




#
# (c) 2021-2022 Giorgio Gonnella, University of Goettingen, Germany
#

import multiplug
import yaml
import os
from sqlalchemy import inspect
from sqlalchemy.orm import Session
from prenacs import plugins_helper, PluginDescription, ComputationReport

class ResultsLoader():

  @staticmethod
  def _different_fields(obj1, obj2):
    result = []
    for c in inspect(obj1).mapper.c:
      v1 = getattr(obj1, c.name)
      v2 = getattr(obj2, c.name)
      if v1 != v2:
        if hasattr(c, "default"):
          if (v1 is None and v2 == c.default.arg) or \
             (v2 is None and v1 == c.default.arg):
            continue
        result.append((c, v1, v2))
    return result

  def _insert_update_or_compare(self, session, klass, newdata,
                                primarykeys, replace, msg):
    newrow = klass(**newdata)
    oldrow = session.get(klass, primarykeys)
    if oldrow:
      if replace:
        session.merge(newrow)
      else:
        diff = self._different_fields(newrow, oldrow)
        if diff:
          raise RuntimeError(f"{msg}\nDifferences:\n"+
          "\n".join([f"  new {c}: {v1}\n  old {c}: {v2}" \
                     for c, v1, v2 in diff]))
    else:
      session.add(newrow)

  def _process_plugin_description(self, session, plugin, replace):
    self._insert_update_or_compare(session, PluginDescription,
        plugins_helper.plugin_metadata_str(plugin),
        [self.plugin.ID, self.plugin.VERSION], replace,
        "Plugin metadata changed, without a version change\n"
        "Please either set the replace-plugin-record option or increase the "+\
        "plugin version number")

  def __init__(self, attribute_value_tables, plugin_fn,
               replace_plugin_record=False, verbose=False):
    self.plugin = multiplug.importer(plugin_fn, verbose=verbose,
                                     **plugins_helper.COMPUTE_PLUGIN_INTERFACE)
    self.connection = attribute_value_tables.connectable
    self.avt = attribute_value_tables
    session = Session(bind=self.connection)
    self._process_plugin_description(session, self.plugin,
                                     replace_plugin_record)
    session.commit()

  def _check_plugin_key(self, report_data):
    for key in ["ID", "VERSION"]:
      report_value = report_data["plugin_"+key.lower()]
      exp_value = getattr(self.plugin, key)
      if report_value != exp_value:
        raise RuntimeError(f"Plugin {key} mismatch\n"+
            f"{key} from the plugin module: {exp_value}\n"+\
            f"{key} from the computation report: {report_value}\n")

  def _process_computation_report(self, report_file, replace):
    session = Session(bind=self.connection)
    with open(report_file) as report:
      report_data = yaml.safe_load(report)
    self._check_plugin_key(report_data)
    uuid = report_data["uuid"]
    self._insert_update_or_compare(session,\
        ComputationReport, report_data, uuid,
        replace, "A computation report with the same ID "+\
            "({uuid}) "+\
            "was already stored in the database\n"
            "Please set the replace-report-record option or "+\
            "use a different computation report ID.")
    session.commit()
    return uuid

  def run(self, results_file, report_file, replace_report_record=False,
          verbose=False):
    if os.stat(results_file).st_size == 0:
      raise RuntimeError("The results file is empty")
    else:
      computation_id = self._process_computation_report(\
                         report_file, replace_report_record)
      self.avt.load_computation(computation_id,
                                self.plugin.OUTPUT,
                                results_file)




#
# (c) 2021-2022 Giorgio Gonnella, University of Goettingen, Germany
#
from .batch_computation import BatchComputation
from .plugin_interface_analyser import PluginInterfaceAnalyser
from .attribute_definitions_manager import AttributeDefinitionsManager
from .dbschema.attribute_definition import AttributeDefinition
from .dbschema.plugin_description import PluginDescription
from .dbschema.computation_report import ComputationReport
from .results_loader import ResultsLoader



#
# (c) 2021-2022 Giorgio Gonnella, University of Goettingen, Germany
#
from prenacs.dbschema.attribute_definition import AttributeDefinition
from prenacs.dbschema.plugin_description import PluginDescription
from prenacs.dbschema.computation_report import ComputationReport

DEFAULT_AVT_PREFIX = "prenacs_attribute_value_t"

def create(connection):
  AttributeDefinition.metadata.create_all(connection)
  PluginDescription.metadata.create_all(connection)
  ComputationReport.metadata.create_all(connection)

def drop(connection, attribute_value_tables=None):
  AttributeDefinition.metadata.drop_all(connection)
  PluginDescription.metadata.drop_all(connection)
  ComputationReport.metadata.drop_all(connection)
  if attribute_value_tables:
    attribute_value_tables.drop_all()



from setuptools import setup, find_packages

def readme():
  with open('README.md') as f:
    return f.read()

import sys
if not sys.version_info[0] == 3:
  sys.exit("Sorry, only Python 3 is supported")

setup(name='prenacs',
      version='1.0',
      description='PRovenance-tracking ENtity "+\
                  "Attribute batch Computation and database Storage system',
      long_description=readme(),
      long_description_content_type="text/markdown",
      url='https://github.com/ggonnella/prenacs',
      keywords="batch computing, database, data provenance",
      author='Giorgio Gonnella',
      author_email='gonnella@zbh.uni-hamburg.de',
      license='ISC',
      # see https://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development :: Libraries',
      ],
      packages=find_packages(),
      scripts=['bin/prenacs-load-results',
               'bin/prenacs-batch-compute',
               'bin/prenacs-check-plugin',
               'bin/prenacs-drop-attribute',
               'bin/prenacs-create-attribute',
               'bin/prenacs-manage-attributes'],
      zip_safe=False,
      test_suite="pytest",
      include_package_data=True,
      tests_require=['pytest', 'pytest-console-scripts'],
    )



#
# (c) 2022, Giorgio Gonnella, University of Goettingen, Germany
#

from schema import Or, And, Use
import sys
import yaml
from prenacs.report import Report

YAMLFILE_VALIDATOR = Or(And(None, Use(lambda n: {})),
                        And(Use(open), Use(lambda fn: yaml.safe_load(fn))))

OUTFILE_OR_STDERR_VALIDATOR = Or(And(None, Use(lambda f: sys.stderr)),
                              Use(lambda f: open(f, "w")))

ARGS_SCHEMA = {"--report": OUTFILE_OR_STDERR_VALIDATOR,
               "--user": Or(str, None),
               "--reason": Or(None, lambda r: r in Report.REASONS),
               "--system": Or(str, None),
               "--params": YAMLFILE_VALIDATOR}

ARGS_DOC = """\
  --report, -r FN   computation report file (default: stderr)
  --user U          user_id for the report (default: getpass.getuser())
  --system S        system_id for the report (default: socket.gethostname())
  --reason R        reason field for the report (default: None)
  --params FNAME    YAML file with additional parameters (default: None)"""

SNAKE_ARGS = {"input": ["--params"], "output": ["--report"],
              "params": ["--user", "--system", "--reason"]}

